@@ -2,15 +2,15 @@ import math
 from enum import Enum, auto
 from typing import List, Optional, Union
 
-from netqasm.lang.ir import BranchLabel, GenericInstr, ICmd, PreSubroutine
+from netqasm.lang.ir import BranchLabel, GenericInstr, ICmd, ProtoSubroutine
 from netqasm.logging.glob import get_netqasm_logger
 from netqasm.sdk.build_types import NVHardwareConfig
-from netqasm.sdk.compiling import NVSubroutineCompiler
 from netqasm.sdk.connection import DebugConnection
 from netqasm.sdk.constraint import ValueAtMostConstraint
 from netqasm.sdk.epr_socket import EPRSocket
 from netqasm.sdk.futures import RegFuture
-from netqasm.sdk.qubit import Qubit
+from netqasm.sdk.qubit import Qubit, QubitMeasureBasis
+from netqasm.sdk.transpile import NVSubroutineTranspiler
 
 logger = get_netqasm_logger()
 
@@ -119,8 +119,8 @@ class PatternMatcher:
             self._cmd_idx += 1
 
 
-class PreSubroutineInspector:
-    def __init__(self, subroutine: PreSubroutine) -> None:
+class ProtoSubroutineInspector:
+    def __init__(self, subroutine: ProtoSubroutine) -> None:
         self._subroutine = subroutine
 
     def contains_instr(self, instr_type: GenericInstr) -> bool:
@@ -130,6 +130,15 @@ class PreSubroutineInspector:
                 return True
         return False
 
+    def find_instr(self, instr_type: GenericInstr) -> List[ICmd]:
+        instrs: List[ICmd] = []
+        for cmd in self._subroutine.commands:
+            if not isinstance(cmd, ICmd):
+                continue
+            if cmd.instruction == instr_type:
+                instrs.append(cmd)
+        return instrs
+
     def match_pattern(
         self, pattern: List[Union[GenericInstr, PatternWildcard]]
     ) -> bool:
@@ -148,7 +157,7 @@ def test_simple():
         subroutine = conn.builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
     assert inspector.contains_instr(GenericInstr.QALLOC)
     assert inspector.contains_instr(GenericInstr.SET)
     assert not inspector.contains_instr(GenericInstr.ROT_X)
@@ -189,7 +198,7 @@ def test_create_epr():
         subroutine = conn.builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
 
     assert inspector.match_pattern(
         [
@@ -229,7 +238,7 @@ def test_branching():
         subroutine = conn.builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
 
     assert inspector.match_pattern(
         [
@@ -261,7 +270,7 @@ def test_loop_context():
         subroutine = conn.builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
     assert inspector.contains_instr(GenericInstr.QALLOC)
     assert inspector.contains_instr(GenericInstr.SET)
     assert not inspector.contains_instr(GenericInstr.ROT_X)
@@ -289,7 +298,7 @@ def test_looping():
         subroutine = conn.builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
 
     assert inspector.match_pattern(
         [
@@ -317,7 +326,7 @@ def test_futures():
         subroutine = conn.builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
 
     assert inspector.match_pattern(
         [
@@ -360,7 +369,7 @@ def test_nested():
         subroutine = conn.builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
 
     assert inspector.match_pattern(
         [
@@ -424,7 +433,44 @@ def test_epr_keep_info():
         subroutine = conn._builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
+    assert inspector.match_pattern(
+        [
+            GenericInstr.CREATE_EPR,
+            GenericInstr.WAIT_ALL,
+            PatternWildcard.ANY_ZERO_OR_MORE,  # Bell corrections
+            GenericInstr.ROT_Z,  # Bell corrections
+            PatternWildcard.ANY_ZERO_OR_MORE,  # Bell corrections
+            GenericInstr.LOAD,
+            GenericInstr.BLT,
+            GenericInstr.SET,
+            GenericInstr.QALLOC,
+            GenericInstr.INIT,
+            PatternWildcard.BRANCH_LABEL,
+        ]
+    )
+
+
+def test_epr_keep_info_no_phi_plus():
+    DebugConnection.node_ids = {
+        "Alice": 0,
+        "Bob": 1,
+    }
+
+    epr_socket = EPRSocket("Bob")
+
+    with DebugConnection("Alice", epr_sockets=[epr_socket]) as conn:
+        eprs, infos = epr_socket.create_keep_with_info(expect_phi_plus=False)
+        epr, info = eprs[0], infos[0]
+
+        _ = epr.measure(store_array=False)
+        with info.generation_duration.if_ge(1337):
+            _ = Qubit(conn)
+
+        subroutine = conn._builder.subrt_pop_pending_subroutine()
+        print(subroutine)
+
+    inspector = ProtoSubroutineInspector(subroutine)
     assert inspector.match_pattern(
         [
             GenericInstr.CREATE_EPR,
@@ -458,7 +504,7 @@ def test_epr_context():
         subroutine = conn._builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
     assert inspector.match_pattern(
         [
             GenericInstr.CREATE_EPR,
@@ -496,7 +542,7 @@ def test_epr_context_future_index():
         subroutine = conn._builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
     assert inspector.match_pattern(
         [
             GenericInstr.CREATE_EPR,
@@ -537,7 +583,7 @@ def test_epr_post():
         subroutine = conn._builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
     assert inspector.match_pattern(
         [
             GenericInstr.CREATE_EPR,
@@ -545,6 +591,9 @@ def test_epr_post():
             PatternWildcard.BRANCH_LABEL,
             PatternWildcard.ANY_ZERO_OR_MORE,
             GenericInstr.WAIT_ALL,
+            PatternWildcard.ANY_ZERO_OR_MORE,  # Bell corrections
+            GenericInstr.ROT_Z,  # Bell corrections
+            PatternWildcard.ANY_ZERO_OR_MORE,  # Bell corrections
             GenericInstr.LOAD,
             GenericInstr.H,
             GenericInstr.ADD,
@@ -569,7 +618,7 @@ def test_try():
         subroutine = conn.builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
     assert inspector.match_pattern(
         [
             GenericInstr.QALLOC,
@@ -594,7 +643,7 @@ def test_loop_until():
         subroutine = conn.builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
     assert inspector.match_pattern(
         [
             PatternWildcard.BRANCH_LABEL,
@@ -629,7 +678,7 @@ def test_epr_min_fidelity_all():
         subroutine = conn._builder.subrt_pop_pending_subroutine()
         print(subroutine)
 
-    inspector = PreSubroutineInspector(subroutine)
+    inspector = ProtoSubroutineInspector(subroutine)
     assert inspector.match_pattern(
         [
             GenericInstr.SET,
@@ -637,6 +686,9 @@ def test_epr_min_fidelity_all():
             GenericInstr.BEQ,
             GenericInstr.CREATE_EPR,
             GenericInstr.WAIT_ALL,
+            PatternWildcard.ANY_ZERO_OR_MORE,  # Bell corrections
+            GenericInstr.ROT_Z,  # Bell corrections
+            PatternWildcard.ANY_ZERO_OR_MORE,  # Bell corrections
             GenericInstr.LOAD,
             GenericInstr.BLT,
             PatternWildcard.ANY_ZERO_OR_MORE,
@@ -784,7 +836,7 @@ def test_bqc_receiver_NV():
         "Alice",
         epr_sockets=[epr_socket],
         hardware_config=NVHardwareConfig(2),
-        compiler=NVSubroutineCompiler,
+        compiler=NVSubroutineTranspiler,
     ) as conn:
         carbon, electron = epr_socket.recv_keep(2)
 
@@ -800,9 +852,9 @@ def test_bqc_receiver_NV():
         carbon.H()
         _ = carbon.measure()
 
-        presubroutine = conn.builder.subrt_pop_pending_subroutine()
-        print(presubroutine)
-        compiled_subroutine = conn.builder.subrt_compile_subroutine(presubroutine)
+        protosubroutine = conn.builder.subrt_pop_pending_subroutine()
+        print(protosubroutine)
+        compiled_subroutine = conn.builder.subrt_compile_subroutine(protosubroutine)
         print(compiled_subroutine)
 
 
@@ -818,7 +870,7 @@ def test_bqc_receiver_NV_min_fidelity():
         "Alice",
         epr_sockets=[epr_socket],
         hardware_config=NVHardwareConfig(2),
-        compiler=NVSubroutineCompiler,
+        compiler=NVSubroutineTranspiler,
     ) as conn:
         carbon, electron = epr_socket.recv_keep(
             2, min_fidelity_all_at_end=80, max_tries=25
@@ -836,8 +888,58 @@ def test_bqc_receiver_NV_min_fidelity():
         carbon.H()
         _ = carbon.measure()
 
+        protosubroutine = conn.builder.subrt_pop_pending_subroutine()
+        print(protosubroutine)
+        compiled_subroutine = conn.builder.subrt_compile_subroutine(protosubroutine)
+        print(compiled_subroutine)
+
+
+def test_measure_Z():
+    with DebugConnection("Alice") as conn:
+        q = Qubit(conn)
+        q.measure(basis=QubitMeasureBasis.Z)
+
         presubroutine = conn.builder.subrt_pop_pending_subroutine()
         print(presubroutine)
+        inspector = ProtoSubroutineInspector(presubroutine)
+        assert inspector.contains_instr(GenericInstr.MEAS)
+
+        compiled_subroutine = conn.builder.subrt_compile_subroutine(presubroutine)
+        print(compiled_subroutine)
+
+
+def test_measure_basis():
+    with DebugConnection("Alice") as conn:
+        q = Qubit(conn)
+        q.measure(basis=QubitMeasureBasis.X)
+
+        presubroutine = conn.builder.subrt_pop_pending_subroutine()
+        print(presubroutine)
+        inspector = ProtoSubroutineInspector(presubroutine)
+        assert inspector.contains_instr(GenericInstr.MEAS_BASIS)
+
+        [meas_basis] = inspector.find_instr(GenericInstr.MEAS_BASIS)
+        # check if rotations are correct
+        assert meas_basis.operands[2:] == [0, 24, 0, 4]  # skip register operands
+
+        compiled_subroutine = conn.builder.subrt_compile_subroutine(presubroutine)
+        print(compiled_subroutine)
+
+
+def test_measure_basis_rotation():
+    with DebugConnection("Alice") as conn:
+        q = Qubit(conn)
+        q.measure(basis_rotations=(3, 4, 5))
+
+        presubroutine = conn.builder.subrt_pop_pending_subroutine()
+        print(presubroutine)
+        inspector = ProtoSubroutineInspector(presubroutine)
+        assert inspector.contains_instr(GenericInstr.MEAS_BASIS)
+
+        [meas_basis] = inspector.find_instr(GenericInstr.MEAS_BASIS)
+        # check if rotations are correct
+        assert meas_basis.operands[2:] == [3, 4, 5, 4]  # skip register operands
+
         compiled_subroutine = conn.builder.subrt_compile_subroutine(presubroutine)
         print(compiled_subroutine)
 
@@ -852,6 +954,7 @@ if __name__ == "__main__":
     test_futures()
     test_nested()
     test_epr_keep_info()
+    test_epr_keep_info_no_phi_plus()
     test_epr_context()
     test_epr_context_future_index()
     test_epr_post()
@@ -863,3 +966,6 @@ if __name__ == "__main__":
     test_create_3_eprs_NV()
     test_bqc_receiver_NV()
     test_bqc_receiver_NV_min_fidelity()
+    test_measure_Z()
+    test_measure_basis()
+    test_measure_basis_rotation()
